"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.main = void 0;
const config_loader_1 = require("./config-loader");
const logger_1 = require("./logger");
const executor_1 = require("./executor");
const launch_menu_1 = require("./launch-menu");
const fs = require("fs");
const path = require("path");
const common_1 = require("./common");
const scripts_1 = require("./scripts");
const package_json_1 = require("./package.json");
const prettyTime = require("pretty-time");
const os = require("os");
const child_process_1 = require("child_process");
const npmScripts = [
    'prepublish',
    'prepare',
    'prepublishOnly',
    'prepack',
    'postpack',
    'publish',
    'postpublish',
    'preinstall',
    'install',
    'postinstall',
    'preuninstall',
    'uninstall',
    'postuninstall',
    'preversion',
    'version',
    'postversion',
    'pretest',
    'test',
    'posttest',
    'prestop',
    'stop',
    'poststop',
    'prestart',
    'start',
    'poststart',
    'prerestart',
    'restart',
    'postrestart',
    'preshrinkwrap',
    'shrinkwrap',
    'postshrinkwrap'
];
function showLoadedFiles(files) {
    for (const file of files) {
        logger_1.Logger.info('Loaded config: ', file);
    }
    logger_1.Logger.info();
}
function showTemplates() {
    const templatePath = path.join(__dirname, 'templates');
    console.log(common_1.Colors.Bold + 'Available templates:' + common_1.Colors.Normal);
    console.log();
    for (const fileName of fs.readdirSync(templatePath)) {
        console.log(fileName);
    }
    console.log();
    console.log(common_1.Colors.Bold + 'Example usage:' + common_1.Colors.Normal + ' npx launch init basic');
}
function copyTemplateFiles(template, directory) {
    const templatePath = path.join(__dirname, 'templates', template);
    console.log(common_1.Colors.Bold + 'Create starter config:' + common_1.Colors.Normal, template);
    console.log();
    if (!fs.existsSync(templatePath) || !fs.statSync(templatePath).isDirectory()) {
        throw new Error('Template not found.');
    }
    for (const fileName of fs.readdirSync(templatePath)) {
        const sourceFile = path.join(templatePath, fileName);
        const targetFile = path.join(directory, fileName);
        if (!fs.existsSync(targetFile)) {
            console.log(common_1.Colors.Bold + 'Creating:' + common_1.Colors.Normal, targetFile.replace(process.cwd() + path.sep, ''));
            fs.copyFileSync(sourceFile, targetFile);
        }
        else {
            console.log(common_1.Colors.Yellow + common_1.Colors.Bold + 'Skipped:' + common_1.Colors.Normal, fileName + ' already exists.');
        }
    }
}
function splitCommand(command) {
    const result = [];
    let last = 0;
    let index = 0;
    while (index < command.length) {
        const value = command.substr(index);
        const semicolon = value.startsWith(';');
        if (value.startsWith('&&') || semicolon) {
            result.push(command.substr(last, index - last).trim() + (semicolon ? ' || true' : ''));
            index++;
            last = index + 1;
        }
        if (value.startsWith('(')) {
            let open = 1;
            while (open > 0 && ++index < command.length) {
                if (command[index] === '(')
                    open++;
                if (command[index] === ')')
                    open--;
            }
        }
        if (value.startsWith('"')) {
            while (++index < command.length && command[index] !== '"')
                ;
        }
        if (value.startsWith("'")) {
            while (++index < command.length && command[index] !== "'")
                ;
        }
        index++;
    }
    result.push(command.substr(last, index - last).trim());
    return result;
}
function checkCleanGit() {
    try {
        const result = (0, child_process_1.execSync)('git status --porcelain', { encoding: 'utf8', stdio: 'pipe' });
        return result.trim().length === 0;
    }
    catch (_a) { }
    return true;
}
function checkMigratePrerequisites(directory, scripts, testmode) {
    const menuFile = path.join(directory, 'launcher-menu.json');
    const configFile = path.join(directory, 'launcher-config.json');
    if (!testmode && !checkCleanGit()) {
        console.log(common_1.Colors.Red + common_1.Colors.Bold + 'Failed:' + common_1.Colors.Normal, 'Repository is not clean. Please commit or stash any changes before updating.');
        return false;
    }
    if (fs.existsSync(menuFile)) {
        console.log(common_1.Colors.Red + common_1.Colors.Bold + 'Failed:' + common_1.Colors.Normal, 'launcher-menu.json already exists.');
        return false;
    }
    if (fs.existsSync(configFile)) {
        console.log(common_1.Colors.Red + common_1.Colors.Bold + 'Failed:' + common_1.Colors.Normal, 'launcher-config.json already exists.');
        return false;
    }
    if (scripts && scripts.start !== undefined && scripts.start !== 'launch') {
        console.log(common_1.Colors.Red + common_1.Colors.Bold + 'Failed:' + common_1.Colors.Normal + ' Remove start script from package.json before running migrate.');
        return false;
    }
    return true;
}
function migrateMenu(scripts) {
    const menuEntries = {
        description: ''
    };
    for (const [key] of Object.entries(scripts)) {
        const entries = key.split(':');
        let currMenu = menuEntries;
        let nextMenu = menuEntries;
        let entry = key;
        for (const item of entries) {
            entry = item;
            currMenu = nextMenu;
            if (currMenu === menuEntries)
                entry += ':...';
            while (typeof currMenu[entry] === 'string')
                entry += ':menu';
            if (nextMenu[entry] === undefined) {
                nextMenu[entry] = {
                    description: ''
                };
            }
            nextMenu = nextMenu[entry];
        }
        if (Object.entries(currMenu[entry]).length > 1)
            entry += ':command';
        currMenu[entry] = key;
    }
    return menuEntries;
}
function migrateScripts(scripts) {
    const sourceScripts = {};
    const targetScripts = {};
    for (const [key, value] of Object.entries(scripts).sort(([key1], [key2]) => key1.localeCompare(key2))) {
        let values = splitCommand(value);
        if (values.length > 1) {
            values = values.map(item => {
                if (item.startsWith('npm run ')) {
                    item = item.trim().replace('npm run ', '');
                    item = item.trim().replace(' || true', '');
                    item = item.trim();
                }
                return item;
            });
            values = values.map(item => {
                if (item.startsWith('cd ')) {
                    item = item.trim().replace('cd ', '');
                    item = item.trim().replace(' || true', '');
                    item = item.trim();
                }
                return item;
            });
            targetScripts[key] = values;
        }
        else {
            targetScripts[key] = value;
        }
        if (npmScripts.includes(key))
            sourceScripts[key] = 'launch';
    }
    return {
        source: sourceScripts,
        target: targetScripts
    };
}
function objectFromEntries(entries) {
    const object = {};
    for (const [name, value] of entries) {
        object[name] = value;
    }
    return object;
}
function parameterize(key, value, preserveParams) {
    const params = key.split(':');
    let index = 0;
    for (const param of params) {
        if (index >= preserveParams) {
            const expression = new RegExp(param, 'g');
            if (value.includes(param)) {
                key = key.replace(expression, '$param' + index);
                value = value.replace(expression, '$param' + index);
            }
        }
        index++;
    }
    return { key, value, params };
}
function getScriptDefinitions(scripts, preserveParams) {
    const definitions = {};
    for (const [key, value] of Object.entries(scripts)) {
        const parameters = parameterize(key, value, preserveParams);
        let definition = definitions[parameters.value];
        if (definition === undefined) {
            definition = {
                keys: [],
                params: []
            };
            definitions[parameters.value] = definition;
        }
        if (!definition.keys.includes(parameters.key))
            definition.keys.push(parameters.key);
        definition.params.push(parameters.params);
    }
    const sorted = Object.entries(definitions).sort(([, definition1], [, definition2]) => definition2.params.length - definition1.params.length);
    return objectFromEntries(sorted);
}
function expandParams(name, command, params) {
    let match = null;
    const length = (params.find(() => true) || { length: 0 }).length;
    const keys = name.split(':');
    for (let index = 0; index < length; index++) {
        if (keys[index].startsWith('$')) {
            const value = [...new Set(params.map(item => item[index]))];
            if (match === null || value.length <= match.value.length) {
                match = {
                    index: index,
                    value: value
                };
            }
        }
    }
    if (match === null)
        return [];
    const expression = new RegExp('\\$param' + match.index, 'g');
    const result = [];
    for (const item of match.value) {
        result.push({
            name: name.replace(expression, item),
            command: command.replace(expression, item)
        });
    }
    return result;
}
function resolveConflicts(scripts, name, command, params) {
    if (scripts[name] === undefined)
        return [{ name, command }];
    const expanded = expandParams(name, command, params);
    const result = [];
    for (const item of expanded) {
        result.push(...resolveConflicts(scripts, item.name, item.command, params));
    }
    return result;
}
function combineScripts(scripts, preserveParams) {
    const definitions = getScriptDefinitions(scripts, preserveParams);
    const combineScripts = {};
    for (const [script, definition] of Object.entries(definitions)) {
        for (const key of definition.keys) {
            const scripts = [];
            scripts.push(...resolveConflicts(combineScripts, key, script, definition.params));
            for (const script of scripts) {
                combineScripts[script.name] = script.command;
            }
        }
    }
    return combineScripts;
}
function migratePackageJson(directory, preserveParams, confirm, testmode) {
    return __awaiter(this, void 0, void 0, function* () {
        const menuFile = path.join(directory, 'launcher-menu.json');
        const configFile = path.join(directory, 'launcher-config.json');
        const packageFile = path.join(directory, 'package.json');
        console.log(common_1.Colors.Bold + 'Migrating: ' + common_1.Colors.Normal + 'package.json');
        console.log();
        const content = JSON.parse(fs.readFileSync(packageFile).toString());
        if (!checkMigratePrerequisites(directory, content.scripts, testmode))
            return;
        const menuEntries = migrateMenu(content.scripts);
        const combinedScripts = combineScripts(content.scripts, preserveParams);
        const scripts = migrateScripts(combinedScripts);
        const targetCount = Object.entries(scripts.target).length;
        const sourceCount = Object.entries(scripts.source).length;
        console.log('Script to migrate:', targetCount - sourceCount);
        console.log('Script to update:', sourceCount + 1);
        console.log();
        scripts.source.start = 'launch';
        content.scripts = scripts.source;
        logger_1.Logger.log('package.json:', content);
        logger_1.Logger.log();
        logger_1.Logger.log('launcher-menu.json:', {
            menu: menuEntries
        });
        logger_1.Logger.log();
        logger_1.Logger.log('launcher-config.json:', {
            scripts: scripts.target
        });
        logger_1.Logger.log();
        let autoValue;
        if (confirm !== undefined)
            autoValue = confirm;
        if (testmode)
            autoValue = true;
        if (yield (0, common_1.confirmPrompt)('Are you sure', autoValue)) {
            console.log();
            console.log(common_1.Colors.Bold + 'Updating:' + common_1.Colors.Normal, packageFile.replace(process.cwd() + path.sep, ''));
            fs.writeFileSync(packageFile, JSON.stringify(content, null, 2) + '\n');
            console.log(common_1.Colors.Bold + 'Creating:' + common_1.Colors.Normal, menuFile.replace(process.cwd() + path.sep, ''));
            fs.writeFileSync(menuFile, JSON.stringify({
                menu: menuEntries
            }, null, 2) + '\n');
            console.log(common_1.Colors.Bold + 'Creating:' + common_1.Colors.Normal, configFile.replace(process.cwd() + path.sep, ''));
            fs.writeFileSync(configFile, JSON.stringify({
                scripts: scripts.target
            }, null, 2) + '\n');
        }
    });
}
function updatePackageJson(directory) {
    const fileName = path.join(directory, 'package.json');
    console.log(common_1.Colors.Bold + 'Updating package.json.' + common_1.Colors.Normal);
    console.log();
    if (!fs.existsSync(fileName)) {
        console.log(common_1.Colors.Bold + 'Update package.json failed:' + common_1.Colors.Normal + ' file not found.');
        return;
    }
    try {
        const buffer = fs.readFileSync(fileName);
        const content = JSON.parse(buffer.toString());
        if (content.scripts && content.scripts.start !== undefined) {
            if (content.scripts.start === 'launch') {
                console.log(common_1.Colors.Yellow + common_1.Colors.Bold + 'Skipped:' + common_1.Colors.Normal + ' start script is up-to-date.');
            }
            else {
                console.log(common_1.Colors.Yellow + common_1.Colors.Bold + 'Skipped:' + common_1.Colors.Normal + ' start script already in use.');
            }
            return;
        }
        if (!content.scripts)
            content.scripts = {};
        content.scripts.start = 'launch';
        fs.writeFileSync(fileName, JSON.stringify(content, null, 2) + '\n');
        console.log(common_1.Colors.Bold + 'Start script of package.json updated.' + common_1.Colors.Normal);
    }
    catch (error) {
        console.log(common_1.Colors.Bold + 'Update package.json failed: ' + common_1.Colors.Normal + error.message);
    }
}
function showHelp() {
    (0, common_1.showArgsHelp)('launch', {
        init: ['', 'Commands:', '  ' + common_1.Colors.Cyan + 'init         ' + common_1.Colors.Normal + '[template] Create starter config files.'],
        list: '  ' + common_1.Colors.Cyan + 'list         ' + common_1.Colors.Normal + '[type] List available launcher scripts.',
        migrate: '  ' + common_1.Colors.Cyan + 'migrate      ' + common_1.Colors.Normal + 'Migrate your package.json scripts.',
        help: '  ' + common_1.Colors.Cyan + 'help         ' + common_1.Colors.Normal + 'Show this help.',
        version: '  ' + common_1.Colors.Cyan + 'version      ' + common_1.Colors.Normal + 'Outputs launcher version.',
        logLevel: ['', 'Options:', '  ' + common_1.Colors.Cyan + 'logLevel=    ' + common_1.Colors.Normal + 'Set log level.'],
        dry: '  ' + common_1.Colors.Cyan + 'dry=         ' + common_1.Colors.Normal + 'Do not execute commands.',
        config: '  ' + common_1.Colors.Cyan + 'config=      ' + common_1.Colors.Normal + 'Merge in an extra config file.',
        confirm: '  ' + common_1.Colors.Cyan + 'confirm=     ' + common_1.Colors.Normal + 'Auto value for confirm conditions.',
        ansi: '  ' + common_1.Colors.Cyan + 'ansi=        ' + common_1.Colors.Normal + 'Enable or disable ansi color output.',
        directory: '  ' + common_1.Colors.Cyan + 'directory=   ' + common_1.Colors.Normal + 'The directory from which configuration files are loaded.',
        menuTimeout: '  ' + common_1.Colors.Cyan + 'menuTimeout= ' + common_1.Colors.Normal + 'Set menu timeout in seconds.',
        params: '  ' + common_1.Colors.Cyan + 'params=      ' + common_1.Colors.Normal + 'Set the number of parameters to preserve.',
        concurrent: '  ' + common_1.Colors.Cyan + 'concurrent=  ' + common_1.Colors.Normal + 'Execute commandline wildcard matches in parallel.',
        limit: '  ' + common_1.Colors.Cyan + 'limit=       ' + common_1.Colors.Normal + 'Limit the number of commands to execute in parallel.'
    });
}
function disableAnsiColors() {
    for (const key of Object.keys(common_1.Colors)) {
        common_1.Colors[key] = '';
    }
}
function getEnviromentValues() {
    const environment = Object.assign({}, process.env);
    for (const [key, value] of Object.entries(common_1.Colors)) {
        environment['launch_style_' + key.toLowerCase()] = value;
    }
    environment.launch_time_start = (0, common_1.formatTime)();
    environment.launch_platform = process.platform;
    environment.launch_version = package_json_1.version;
    delete environment.launch_time_current;
    delete environment.launch_time_elapsed;
    return environment;
}
function getLaunchSetting(settings, prefix = 'launch_setting_') {
    const result = {
        values: {},
        arrays: {}
    };
    for (const [key, value] of Object.entries(settings)) {
        if (value instanceof Array) {
            const name = prefix + key;
            result.arrays[name] = [];
            for (const item of value) {
                if (typeof item !== 'object') {
                    result.arrays[name].push({
                        ['_']: item
                    });
                    continue;
                }
                const settings = getLaunchSetting(item, '_');
                result.arrays[name].push(settings.values);
            }
            continue;
        }
        if (typeof value === 'object') {
            const settings = getLaunchSetting(value, prefix + key + '_');
            result.values = Object.assign(Object.assign({}, result.values), settings.values);
            result.arrays = Object.assign(Object.assign({}, result.arrays), settings.arrays);
            continue;
        }
        result.values[prefix + key.toLowerCase()] = value;
    }
    return result;
}
function getRemaining(args) {
    const index = args.indexOf('--');
    if (index === -1)
        return [];
    return args.slice(index + 1);
}
function getMenuScripts(menu, result = []) {
    for (const [key, value] of Object.entries(menu)) {
        if (key === 'description')
            continue;
        if (key === 'separator')
            continue;
        if (typeof value === 'string') {
            if (value.includes(' '))
                continue;
            result.push(value);
        }
        else {
            getMenuScripts(value, result);
        }
    }
    return result;
}
function main(lifecycleEvent, processArgv, npmConfigArgv, testmode = false) {
    return __awaiter(this, void 0, void 0, function* () {
        let exitCode = 1;
        let startTime = process.hrtime();
        try {
            const commandArgs = npmConfigArgv || [];
            const argsString = processArgv.slice(2, processArgv.length - commandArgs.length);
            const launchArgs = (0, common_1.parseArgs)(argsString, {
                arguments: {
                    logLevel: undefined,
                    dry: undefined,
                    init: false,
                    list: false,
                    migrate: false,
                    confirm: undefined,
                    help: false,
                    version: false,
                    config: null,
                    script: null,
                    ansi: true,
                    directory: process.cwd(),
                    menuTimeout: undefined,
                    params: undefined,
                    concurrent: false,
                    limit: undefined
                },
                optionals: [],
                unknowns: []
            });
            launchArgs.arguments.directory = path.join(launchArgs.arguments.directory); // remove starting ./
            const configLoad = config_loader_1.Config.load(launchArgs.arguments.directory);
            let config = configLoad.config;
            let interactive = false;
            if (launchArgs.arguments.logLevel === undefined)
                launchArgs.arguments.logLevel = config.options.logLevel;
            if (launchArgs.arguments.dry === undefined)
                launchArgs.arguments.dry = config.options.dry;
            if (launchArgs.arguments.menuTimeout === undefined)
                launchArgs.arguments.menuTimeout = config.options.menu.timeout;
            if (argsString.includes('--params') && launchArgs.arguments.params === undefined)
                launchArgs.arguments.params = 1;
            if (launchArgs.arguments.params === undefined)
                launchArgs.arguments.params = Number.MAX_SAFE_INTEGER;
            if (launchArgs.arguments.limit === undefined)
                launchArgs.arguments.limit = config.options.limit;
            if (launchArgs.arguments.limit === 0)
                launchArgs.arguments.limit = os.cpus().length;
            if (launchArgs.arguments.limit < 1)
                launchArgs.arguments.limit = 1;
            if (launchArgs.arguments.dry && launchArgs.arguments.logLevel < 1)
                launchArgs.arguments.logLevel = 1;
            logger_1.Logger.level = launchArgs.arguments.logLevel;
            if (launchArgs.arguments.config) {
                const fileName = path.join(launchArgs.arguments.directory, launchArgs.arguments.config);
                config = config.merge(fileName);
                configLoad.files.push(fileName);
            }
            const shell = config_loader_1.Config.evaluateShellOption(config.options.script.shell, true);
            if (!launchArgs.arguments.ansi)
                disableAnsiColors();
            if (process.platform === 'win32')
                common_1.Colors.Dim = '\x1b[90m';
            const settings = getLaunchSetting(config.settings);
            const environment = Object.assign(Object.assign({}, getEnviromentValues()), settings.values);
            if (testmode)
                environment.launch_time_start = (0, common_1.formatTime)(new Date('2019-09-16T12:33:20.628').getTime(), 0);
            showLoadedFiles(configLoad.files);
            let launchScript = lifecycleEvent ? [lifecycleEvent] : [];
            let scriptId = '';
            if (launchArgs.unknowns.length === 0) {
                if (lifecycleEvent === 'start') {
                    launchScript = commandArgs[0] ? [commandArgs[0]] : [];
                    scriptId = commandArgs.shift();
                }
            }
            else {
                launchScript = launchArgs.unknowns;
            }
            logger_1.Logger.debug('Config: ', (0, common_1.stringify)(config));
            logger_1.Logger.info(common_1.Colors.Bold + 'Date              :', environment.launch_time_start + common_1.Colors.Normal);
            logger_1.Logger.info('Version           :', package_json_1.version);
            logger_1.Logger.info('Lifecycle event   :', lifecycleEvent);
            logger_1.Logger.info('Launch script     :', launchScript);
            logger_1.Logger.debug('Process platform  :', process.platform);
            logger_1.Logger.debug('Script shell      :', shell);
            if (logger_1.Logger.level > 2) {
                logger_1.Logger.info('Launch arguments  :', launchArgs);
            }
            else {
                logger_1.Logger.info('Launch arguments  :', argsString);
            }
            if (Object.entries(config.scripts.scripts).length === 0) {
                logger_1.Logger.info();
                logger_1.Logger.info('Warning: No launcher scripts loaded.');
                logger_1.Logger.info();
            }
            if (launchArgs.arguments.version) {
                console.log(package_json_1.version);
                logger_1.Logger.log();
                exitCode = 0;
                return;
            }
            if (launchArgs.arguments.help) {
                showHelp();
                logger_1.Logger.log();
                exitCode = 0;
                return;
            }
            if (launchArgs.arguments.init) {
                const template = launchArgs.optionals[0];
                if (!template) {
                    showTemplates();
                    return;
                }
                copyTemplateFiles(template, launchArgs.arguments.directory);
                console.log();
                updatePackageJson(launchArgs.arguments.directory);
                logger_1.Logger.log();
                exitCode = 0;
                return;
            }
            if (launchArgs.arguments.list) {
                if (launchArgs.optionals[0] === 'script') {
                    const choices = Object.keys(configLoad.config.scripts.scripts).sort();
                    const unique = [...new Set(choices)];
                    for (const item of unique) {
                        console.log(item);
                    }
                    return;
                }
                if (launchArgs.optionals[0] === 'menu') {
                    const choices = getMenuScripts(configLoad.config.menu).sort();
                    const unique = [...new Set(choices)];
                    for (const item of unique) {
                        console.log(item);
                    }
                    return;
                }
                if (launchArgs.optionals.length === 0 || launchArgs.optionals[0] === 'complete') {
                    const scripts = Object.keys(configLoad.config.scripts.scripts).filter(item => !item.includes('$'));
                    const menu = getMenuScripts(configLoad.config.menu).filter(item => !scripts.includes(item));
                    const choices = [...menu, ...scripts].sort();
                    const unique = [...new Set(choices)];
                    for (const item of unique) {
                        console.log(item);
                    }
                    return;
                }
                console.error('List option not supported: ' + launchArgs.optionals);
                console.error();
                console.error('Use: script, menu or complete');
                throw new Error();
            }
            if (launchArgs.arguments.migrate) {
                yield migratePackageJson(launchArgs.arguments.directory, launchArgs.arguments.params, launchArgs.arguments.confirm, testmode);
                logger_1.Logger.log();
                exitCode = 0;
                return;
            }
            commandArgs.unshift(...getRemaining(argsString));
            if (scriptId)
                commandArgs.unshift(scriptId);
            const scripts = config.scripts.find(...launchScript);
            const defaultScript = config.scripts.find('start');
            if (scripts.length === 0) {
                if (launchScript[0] === 'menu') {
                    interactive = true;
                    launchScript = [];
                    commandArgs.shift();
                    defaultScript.length = 0;
                }
                if (defaultScript.length > 0) {
                    scripts.push(...defaultScript);
                    launchScript = ['start'];
                }
            }
            if (launchScript.length === 0) {
                logger_1.Logger.info();
                const result = yield (0, launch_menu_1.launchMenu)(environment, settings, config, commandArgs, interactive, launchArgs.arguments.menuTimeout, config.options.menu.confirm, launchArgs.arguments.confirm, launchArgs.arguments.limit, launchArgs.arguments.dry, testmode);
                startTime = result.startTime;
                exitCode = result.exitCode;
                return;
            }
            const scriptInfo = scripts_1.Scripts.select(scripts);
            if (!scriptInfo)
                throw new Error('Cannot start launch script ' + JSON.stringify(launchScript, null, 0) + ': No such script available.');
            if (scriptInfo.multiple && launchArgs.arguments.concurrent) {
                scriptInfo.script = {
                    concurrent: scriptInfo.script
                };
            }
            if (launchArgs.unknowns.length === 0 && lifecycleEvent === 'start') {
                commandArgs[0] = scripts_1.Scripts.parse(launchScript[0]).command;
            }
            else {
                commandArgs.unshift(scripts_1.Scripts.parse(launchScript[0]).command);
            }
            if (!scriptInfo.name)
                scriptInfo.name = launchScript[0];
            scriptInfo.arguments = commandArgs;
            logger_1.Logger.info();
            const executor = new executor_1.Executor(shell, environment, settings, config.scripts, config.options.glob, launchArgs.arguments.confirm, launchArgs.arguments.limit, launchArgs.arguments.dry, testmode);
            startTime = executor.startTime;
            exitCode = yield executor.execute(scriptInfo);
        }
        catch (error) {
            let message = `${error}`;
            if (error.message)
                message = error.message;
            if (message !== 'false' && message !== 'Error')
                logger_1.Logger.error(message);
        }
        finally {
            let timespan = process.hrtime(startTime);
            if (logger_1.Logger.level < 2)
                logger_1.Logger.info('');
            logger_1.Logger.info('ExitCode:', exitCode);
            if (testmode)
                timespan = [0, 237 * 1000 * 1000];
            logger_1.Logger.info('Elapsed: ' + prettyTime(timespan, 'ms'));
            if (!testmode)
                process.exit(exitCode);
        }
    });
}
exports.main = main;
